'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

const Symbol_observable = (() => (typeof Symbol === 'function' && Symbol.observable) || '@@observable')();

let ctx_cc = null;
let ctx_cn = null;
let ctx_cln = null;
let ctx_lr = null;
function setData(data, value) {
    if (!ct) {
        return batch(() => setData(data, value));
    }
    data.value = value;
    if (data.flags & 8 /* Changed */) {
        return;
    }
    data.flags |= 8 /* Changed */;
    ct.addDirty(data);
}
/**
 * Get the value of a node _in many contexts_. Full of dirty side-effects.
 * @param data
 * @returns current value, or previous value if circularly referenced
 * @throws User-land errors, DEFER_COMPUTATION
 */
function accessValue(data) {
    if (data.flags & 16 /* Estimating */) {
        /**
         * In fact we could allow circular referencing, by returning the latest value.
         * However this causes a memo node to _hold a state_.
         * Thus a memo expression is _impure_: same inputs, changed result.
         */
        throw new ReferenceError('A circular reference occurred.'); // [EXIT 5]
    }
    if (data.flags & 32 /* MarkForCheck */ && ct !== null && !ct.flushing) {
        ct.flush();
    }
    if (ctx_cc !== null) {
        logDependency(data);
        // if propagating and current are still markForCheck
        if (data.flags & 32 /* MarkForCheck */) {
            // currently if MarkForCheck it must be in propagate phase
            throw DEFER_COMPUTATION; // [EXIT 4]
        }
    }
    if (data.flags & 1024 /* StaleMemo */) {
        data.flags &= ~1024 /* StaleMemo */;
        estimate(data); // maybe [EXIT 3]
    }
    if (data.flags & 2048 /* HasError */) {
        throw data.value; // [EXIT 2]
    }
    return data.value; // current value [EXIT 1]
}
function estimate(data) {
    data.flags |= 16 /* Estimating */;
    const s_ctx_cc = ctx_cc, s_ctx_cn = ctx_cn, s_ctx_cln = ctx_cln, s_ctx_lr = ctx_lr;
    (ctx_cc = data), (ctx_lr = null), (ctx_cln = null);
    try {
        data.value = evaluate(data, data.c, s_ctx_cc);
        data.flags &= ~2048 /* HasError */;
    }
    catch (e) {
        if (e === DEFER_COMPUTATION) {
            // it will be propagated later
            // _must mark dirty_ because the source may be unchanged but this node must be recomputed.
            data.flags |= 32 /* MarkForCheck */ | 4 /* Dirty */;
            data.dc++;
            throw e; // [EXIT 3]
        }
        data.flags |= 2048 /* HasError */;
        data.value = e;
    }
    finally {
        (ctx_lr = s_ctx_lr), (ctx_cln = s_ctx_cln), (ctx_cn = s_ctx_cn);
        data.flags &= ~16 /* Estimating */;
    }
}
/**
 * Assure ctx_cc is not null
 * @param accessing
 * @returns
 */
function logDependency(accessing) {
    if (ctx_lr) {
        if (ctx_lr === accessing) {
            return;
        }
    }
    ctx_lr = accessing;
    if (ctx_cn === null) {
        insertNewSource(ctx_cc, accessing);
    }
    else {
        if (ctx_cn.source !== accessing) {
            ctx_cln = ctx_cc.ls;
            ctx_cc.ls = ctx_cn.prev;
            ctx_cc.ls.next = null; // new last_source
            ctx_cn.prev = null; //
            ctx_cn = null;
            insertNewSource(ctx_cc, accessing);
        }
        else {
            ctx_cn = ctx_cn.next;
        }
    }
}
const NO_ERROR = {};
/**
 * Compute the latest value of a memo node. And track (refresh) dependencies. (insert then clean-up)
 * @param observer
 * @param expr
 * @returns Computed latest vaue
 * @throws User-land errors, DEFER_COMPUTATION
 */
function evaluate(observer, expr, restore_ctx_cc) {
    ctx_cn = (ctx_cc = observer).fs;
    let calculatedValue, error = NO_ERROR;
    try {
        calculatedValue = expr();
    }
    catch (e) {
        error = e;
    }
    ctx_lr = null;
    if (ctx_cn !== null) {
        // collected source num is less than expected
        // but it's fine as we remove the extra sources.
        cleanupSources(observer.ls, ctx_cn.prev);
        observer.ls = ctx_cn.prev;
        if (observer.ls) {
            observer.ls.next = null;
        }
        else {
            observer.fs = null; // no source. will not update. (but not stale.)
        }
        ctx_cn = null;
    }
    else if (ctx_cln !== null) {
        // why deferred: to not clean a node so early
        cleanupSources(ctx_cln, null);
        ctx_cln = null;
    }
    ctx_cc = restore_ctx_cc;
    if (error !== NO_ERROR) {
        throw error;
    }
    return calculatedValue;
}
function cleanupMemo(memo) {
    if (memo.ls === null) {
        return;
    }
    cleanupSources(memo.ls, null);
    memo.fs = null;
    memo.ls = null;
}
function cleanupSources(last, until) {
    do {
        const { source, observer_ref: { next, prev }, } = last; /* no first source so do nothing */
        // last.observer_ref.observer = null; // not necessary (but might be gc friendly)
        (next ? (next.prev = prev) : (source.lo = prev))
            ? (prev.next = next)
            : null;
        if (source.flags & 2 /* Memo */ && source.lo === null) {
            // it's last observer // but propagation might tackle this.
            cleanupMemo(source);
            source.flags |= 1024 /* StaleMemo */;
        }
    } while ((last = last.prev) !== until);
}
function propagate(stack) {
    while (stack.length > 0) {
        const data = stack.pop();
        const sdf = data.flags;
        const dirtyIfChanged = (sdf & 8 /* Changed */) >> 1; // 8>>1 = 4
        data.flags &= ~8 /* Changed */;
        let lo = data.lo;
        while (lo !== null) {
            const observer = lo.observer;
            lo = lo.prev;
            observer.flags |= dirtyIfChanged; // changes -> dirty
            observer.dc--;
            if (observer.dc === 0) {
                const scf = (observer.flags &= ~32 /* MarkForCheck */);
                if (scf & 256 /* Effect */) {
                    if (scf & 4 /* Dirty */) {
                        observer.flags &= ~4 /* Dirty */;
                        if (~scf & 64 /* EffectWillCommit */) {
                            ct.effects.push(observer);
                            observer.flags |= 64 /* EffectWillCommit */;
                        }
                    }
                }
                else if (observer.lo === null) {
                    // it's a memo but no one observing.
                    cleanupMemo(observer);
                    observer.flags &= ~4 /* Dirty */;
                    observer.flags |= 1024 /* StaleMemo */;
                }
                else {
                    // it's a memo
                    stack.push(observer); // push()
                    if (scf & 4 /* Dirty */) {
                        observer.flags &= ~4 /* Dirty */;
                        // if propagate receives array of data then the if is not necessary
                        try {
                            const currentValue = evaluate(observer, observer.c, null);
                            // compare value , if changed, mark as changed
                            if (currentValue !== observer.value ||
                                observer.flags & 2048 /* HasError */) {
                                observer.flags &= ~2048 /* HasError */;
                                observer.value = currentValue;
                                observer.flags |= 8 /* Changed */;
                            }
                        }
                        catch (e) {
                            if (e === DEFER_COMPUTATION) {
                                // computation deferred as a upstream dependency is waiting for propagation.
                                stack.pop(); // pop()
                                observer.flags |= 4 /* Dirty */ | 32 /* MarkForCheck */;
                                observer.dc++;
                            }
                            else {
                                // TODO: should check error equality?
                                observer.value = e;
                                observer.flags |= 8 /* Changed */ | 2048 /* HasError */;
                            }
                        }
                    }
                }
            }
        }
    }
}
function markObserversForCheck(stack) {
    while (stack.length > 0) {
        const node = stack.pop();
        let lo = node.lo;
        while (lo !== null) {
            const observer = lo.observer;
            lo = lo.prev;
            observer.dc++;
            if (observer.flags & 32 /* MarkForCheck */) {
                continue;
            }
            observer.flags |= 32 /* MarkForCheck */;
            if (~observer.flags & 256 /* Effect */) {
                stack.push(observer);
            }
        }
    }
}
let ct = null;
class Transaction {
    constructor() {
        this.dirty = [];
        this.effects = [];
        this.flushing = false;
    }
    start(fn) {
        const stored = ct;
        ct = this;
        let retValue = null;
        try {
            retValue = fn();
            this.flush();
        }
        finally {
            ct = stored;
        }
        this.commit();
        return retValue;
    }
    flush() {
        this.flushing = true;
        propagate(this.dirty);
        this.flushing = false;
    }
    commit() {
        const effects = this.effects;
        while (effects.length) {
            const x = effects.pop();
            x.flags &= ~64 /* EffectWillCommit */;
            /**
             * should catch error here?
             * no. and unrecoverable.
             */
            x.c();
        }
    }
    addDirty(data) {
        markObserversForCheck([data]);
        this.dirty.push(data);
    }
}
const nt = new Transaction();
function batch(fn) {
    if (ct !== null) {
        return fn();
    }
    return nt.start(fn);
}
function insertNewSource(observer, source) {
    observer.ls = observer.ls
        ? (observer.ls.next = {
            prev: observer.ls,
            next: null,
            source,
            observer_ref: insertNewObserver(source, observer),
        })
        : (observer.fs = {
            prev: null,
            next: null,
            source,
            observer_ref: insertNewObserver(source, observer),
        });
}
function insertNewObserver(source, observer) {
    return (source.lo = source.lo
        ? (source.lo.next = {
            prev: source.lo,
            next: null,
            observer,
        })
        : { prev: null, next: null, observer });
}
function createReaction(onCommit) {
    return {
        flags: 256 /* Effect */,
        lo: null,
        value: null,
        dc: 0,
        fs: null,
        ls: null,
        c: onCommit,
    };
}
function executeReaction(reaction, fn) {
    return evaluate(reaction, fn, null);
}
function createData(value) {
    return {
        flags: 1 /* Data */,
        lo: null,
        value: value,
    };
}
function createMemo(fn) {
    return {
        flags: 2 /* Memo */ | 1024 /* StaleMemo */,
        lo: null,
        value: null,
        dc: 0,
        fs: null,
        ls: null,
        c: fn,
    };
}
function __current_collecting() {
    return ctx_cc;
}
function __current_transaction() {
    return ct;
}
/**
 * Only for _extra dependencies added_
 */
const DEFER_COMPUTATION = {};

class Cell {
    constructor(internal) {
        this.internal = internal;
    }
    static of(initial) {
        return new Cell(createData(initial));
    }
    get value() {
        return accessValue(this.internal);
    }
    get error() {
        try {
            accessValue(this.internal);
            return null;
        }
        catch (e) {
            return e;
        }
    }
    [Symbol_observable]() {
        return this;
    }
    /**
     * @deprecated Should be only used for interop.
     */
    subscribe(next, error, complete) {
        if (typeof next === 'function') {
            return new CellSubscription(this, {
                next,
                error,
                complete,
            });
        }
        else if (typeof next == 'object') {
            return new CellSubscription(this, next);
        }
        throw new TypeError('');
    }
}
class CellSubscription {
    constructor(cell, observer) {
        const callback = () => {
            let value, isError = false;
            try {
                value = this.reaction.execute(() => cell.value);
            }
            catch (e) {
                value = e;
                isError = true;
            }
            if (isError) {
                observer.closed || observer.error?.(value);
            }
            else {
                observer.closed || observer.next?.(value);
            }
        };
        this.reaction = new Reaction(callback);
        callback();
    }
    unsubscribe() {
        this.reaction.dispose();
    }
}
class Reaction {
    constructor(callback) {
        this.internal = createReaction(callback);
    }
    execute(expr) {
        return executeReaction(this.internal, expr);
    }
    dispose() {
        cleanupMemo(this.internal);
    }
}
function mutable(initialValue) {
    const internal = createData(initialValue);
    return [
        new Cell(internal),
        (v) => {
            if (v instanceof Function) {
                setData(internal, v(internal.value));
            }
            else {
                setData(internal, v);
            }
        },
    ];
}
function mutValue(initialValue) {
    const internal = createData(initialValue);
    return [
        new Cell(internal),
        (v) => {
            setData(internal, v);
        },
    ];
}
function computed(expr) {
    return new Cell(createMemo(expr));
}

class SuspendedCell extends Cell {
    constructor(expr, fallback) {
        super(createMemo(() => {
            try {
                accessValue(notifier);
                this.isSuspending = true;
                const value = expr();
                this.isSuspending = false;
                return value;
            }
            catch (e) {
                if (DEFER_SUSPENSION === e) {
                    return fallback;
                }
                if (e instanceof Promise) {
                    e.then(() => {
                        if (e !== awaiting) {
                            return;
                        }
                        awaiting = null;
                        enqueueJob(() => {
                            setData(notifier, notifier.value + 1);
                        });
                    }, (error) => {
                        if (e !== awaiting) {
                            return;
                        }
                        awaiting = null;
                        enqueueJob(() => {
                            // it's definitely in a transaction
                            this.internal.flags |= 2048 /* HasError */ | 8 /* Changed */;
                            this.internal.value = error;
                            __current_transaction().addDirty(this.internal);
                        });
                    });
                    awaiting = e;
                    return fallback;
                }
                this.isSuspending = false; // throw user-land error, or DEFER_COMPUTATION
                throw e;
            }
        }));
        this.isSuspending = false;
        let awaiting = null;
        const notifier = createData(0);
    }
    read() {
        const value = this.value; // log access
        if (this.isSuspending) {
            throw DEFER_SUSPENSION;
        }
        return value;
    }
}
function suspended(expr, fallback) {
    return new SuspendedCell(expr, fallback);
}
const DEFER_SUSPENSION = {
    toString: /* istanbul ignore next */ () => '[Cell SUSPENDING]',
};
const queue = [];
function enqueueJob(job) {
    if (queue.length === 0) {
        enqueueTask(() => {
            batch(() => {
                while (queue.length) {
                    queue.pop()();
                }
            });
        });
    }
    queue.push(job);
}
const enqueueTask = (() => {
    if (typeof setImmediate === 'function') {
        return setImmediate;
    } /* istanbul ignore next */
    if (typeof MessageChannel !== 'undefined') {
        const channel = new MessageChannel();
        return (callback) => {
            channel.port1.onmessage = callback;
            channel.port2.postMessage(undefined);
        };
    } /* istanbul ignore next */
    return (callback) => setTimeout(callback, 0);
})();

function mutArray(initial) {
    let array = initial;
    const internal = createData(array);
    return [
        new Cell(internal),
        {
            set length(value) {
                array = Array.from(array);
                array.length = value;
                setData(internal, array);
                return;
            },
            get length() {
                return array.length;
            },
            setAt(index, value) {
                array[index] = value;
                setData(internal, array);
                return;
            },
            push(...items) {
                const ret = array.push(...items);
                setData(internal, array);
                return ret;
            },
            pop() {
                const ret = array.pop();
                setData(internal, array);
                return ret;
            },
            shift() {
                const ret = array.shift();
                setData(internal, array);
                return ret;
            },
            unshift(...items) {
                const ret = array.unshift(...items);
                setData(internal, array);
                return ret;
            },
            splice(start, deleteCount) {
                const ret = array.splice(start, deleteCount);
                setData(internal, array);
                return ret;
            },
            reverse() {
                const ret = array.reverse();
                setData(internal, array);
                return ret;
            },
        },
    ];
}
function mutSet(initial) {
    let set = new Set(initial);
    const internal = createData(set);
    return [
        new Cell(internal),
        {
            add(value) {
                set = new Set(set);
                set.add(value);
                setData(internal, set);
                return this;
            },
            delete(value) {
                set = new Set(set);
                const ret = set.delete(value);
                setData(internal, set);
                return ret;
            },
            clear() {
                set = new Set();
                setData(internal, set);
            },
        },
    ];
}
function mutMap(initial) {
    let map = new Map(initial ? [...initial] : []);
    const internal = createData(map);
    return [
        new Cell(internal),
        {
            clear() {
                map = new Map();
                setData(internal, map);
            },
            delete(key) {
                map = new Map(map);
                const ret = map.delete(key);
                setData(internal, map);
                return ret;
            },
            set(key, value) {
                map = new Map(map);
                map.set(key, value);
                setData(internal, map);
                return this;
            },
        },
    ];
}

const noop = () => { };
function doCleanup(cleanup) {
    if (typeof cleanup === 'function') {
        cleanup();
    }
    else if (typeof cleanup === 'object') {
        if ('cancel' in cleanup) {
            cleanup.cancel();
        }
        else if ('unsubscribe' in cleanup) {
            cleanup.unsubscribe();
        }
        else if ('dispose' in cleanup) {
            cleanup.dispose();
        }
    }
}

class EventStream {
    constructor(producer) {
        this.producer = producer;
        this.disposeProducer = null;
        this.head = null;
        this.tail = null;
        this.propagating = false;
    }
    *[Symbol.iterator]() {
        return yield (resolve) => {
            const unsub = this.listen((value) => {
                resolve(value);
                unsub(); // 
            });
            return unsub;
        };
    }
    __internal_subscribe(handler) {
        const node = {
            next: null,
            prev: this.tail,
            handler: handler,
        };
        if (!this.head) {
            this.head = node;
            this.tail = node;
            // activate
            this.disposeProducer = this.producer((s) => this.next(s));
        }
        else {
            this.tail.next = node;
            this.tail = node;
        }
        const subscription = () => {
            if (node.prev) {
                //node is head
                node.prev.next = node.next;
            }
            else {
                this.head = node.next;
            }
            if (node.next) {
                //node is tail
                node.next.prev = node.prev;
            }
            else {
                this.tail = node.prev;
            }
            if (!this.head) {
                // deactivate
                doCleanup(this.disposeProducer);
            }
        };
        return subscription;
    }
    next(value) {
        if (this.propagating) {
            console.error(`A loop occured.`); // TODO: detailed error message.
            return;
        }
        this.propagating = true;
        let p = this.head;
        const handlers = [];
        while (p !== null) {
            handlers.push(p.handler);
            p = p.next;
        }
        // linked list is mutable data structure thus a immutable snapshot of subscriptions is required.
        // then you can subscribe or unsubscirbe inside a handler.
        while (handlers.length) {
            handlers.pop()(value);
        }
        this.propagating = false;
    }
    listen(handler) {
        let lastDisposer = undefined;
        const unsubscribe = this.__internal_subscribe((value) => {
            doCleanup(lastDisposer);
            lastDisposer = handler(value);
        });
        return () => {
            doCleanup(lastDisposer);
            unsubscribe();
        };
    }
    transform(transformFn) {
        return new EventStream((next) => {
            return this.__internal_subscribe((v) => next(transformFn(v)));
        });
    }
    filter(filterFn) {
        return new EventStream((next) => {
            return this.__internal_subscribe((v) => filterFn(v) ?? next(v));
        });
    }
    static create() {
        const stream = new EventStream(() => {
            return () => { };
        });
        return [stream, (payload) => batch(() => stream.next(payload))];
    }
}
function stream() {
    return EventStream.create();
}
function never() {
    const [stream] = EventStream.create();
    return stream;
}
function merged(array) {
    return new EventStream((next) => {
        const listeners = array.map((stream) => stream.listen(next));
        return () => listeners.forEach((x) => x());
    });
}

const END = {};
function eventReaderBase(from, until) {
    let closed = false;
    const bufferQueue = [];
    let continuation = null;
    const stopUntilListener = until.listen((next) => {
        if (!closed) {
            dispose();
        }
    });
    const stopFromListener = from.listen((next) => {
        if (continuation) {
            continuation(next);
        }
        else {
            bufferQueue.push(next);
        }
    });
    function dispose() {
        if (!closed) {
            stopUntilListener();
            stopFromListener();
            if (continuation) {
                continuation(END);
                continuation = null;
            }
            closed = true;
        }
    }
    const next = function* () {
        /* istanbul ignore if */
        if (continuation !== null) {
            throw new Error(`There exists a Task waiting for this channel already.`);
        }
        if (closed) {
            throw new Error(`Closed`);
        }
        if (bufferQueue.length > 0) {
            return bufferQueue.shift();
        }
        return (yield (resolve, reject) => {
            continuation = (value) => {
                if (value === END) {
                    reject(new Error(`Closed`));
                    return;
                }
                continuation = null;
                resolve(value);
            };
            return () => {
                continuation = null;
            };
        });
    };
    const hasNext = function* () {
        /* istanbul ignore if */
        if (continuation !== null) {
            throw new Error(`There exists a Task waiting for this channel already.`);
        }
        if (closed) {
            return false;
        }
        if (bufferQueue.length > 0) {
            return true;
        }
        return yield (resolve) => {
            continuation = (value) => {
                if (value === END) {
                    continuation = null;
                    resolve(false);
                    return;
                }
                bufferQueue.push(value);
                continuation = null;
                resolve(true);
            };
            return () => {
                continuation = null;
            };
        };
    };
    return {
        next,
        hasNext,
        dispose,
    };
}
function readEvents(props) {
    return eventReaderBase(props.from, props.until);
}

class CanceledError extends Error {
    constructor(message) {
        super(message ?? 'Operation has been canceled');
        this.name = 'CanceledError';
    }
}
let $$CANT_ACCESS_CHILD_PROP_INSIDE_SUPER = null;
class CancellablePromise extends Promise {
    constructor(executor) {
        super((resolve, reject) => {
            let settled = false;
            // disposer is guaranteed to be not executed after: 1.fulfilled 2.rejected 3.disposed once(which is also reject)
            // resolve in dispose? I think it's valid. And default behavior is reject after disposer executed, but user
            // can reject/resolve in disposer and default behavior is ignored.
            const dispose = executor((v) => {
                resolve(v);
                settled = true;
            }, (e) => {
                reject(e);
                settled = true;
            });
            $$CANT_ACCESS_CHILD_PROP_INSIDE_SUPER = () => {
                if (settled) {
                    return;
                }
                if (dispose) {
                    doCleanup(dispose); // might become settled.
                    settled = true;
                    reject(new CanceledError());
                }
            };
        });
        this.cancel = $$CANT_ACCESS_CHILD_PROP_INSIDE_SUPER;
    }
    *[Symbol.iterator]() {
        return yield (resolve, reject) => {
            this.then(resolve, reject);
            return () => this.cancel();
        };
    }
}

function executeRunnable(runnable, onsuccess, onfailure) {
    let currentDisposer = null;
    const iter = runnable[Symbol.iterator]();
    let settled = false;
    function takeControl(yieldedObject) {
        try {
            if (typeof yieldedObject === 'function') {
                //bind sideEffect
                let synchronous = true; // closure callback
                let innerSettled = false;
                let synchronousResult = null;
                function tryResolve(value) {
                    if (innerSettled)
                        return;
                    innerSettled = true;
                    if (synchronous) {
                        synchronousResult = {
                            $$RESOLVE: true,
                            $$VALUE: value,
                        };
                    }
                    else {
                        resumeTask(value);
                    }
                }
                function tryReject(error) {
                    if (innerSettled)
                        return;
                    innerSettled = true;
                    if (synchronous) {
                        synchronousResult = {
                            $$REJECT: true,
                            $$VALUE: error,
                        };
                    }
                    else {
                        throwTask(error);
                    }
                }
                const maybeDisposer = yieldedObject(tryResolve, tryReject); //might raise an error.
                if (synchronousResult !== null) {
                    throw synchronousResult;
                }
                currentDisposer = () => {
                    // if (innerSettled) return; // no need this line: currentDisposer
                    // is guaranteed to execute only once.
                    doCleanup(maybeDisposer); // maybe raise an error.
                    tryReject(new CanceledError());
                };
                synchronous = false;
            }
            else {
                throw new TypeError('Invalid object yielded. Are you missing an asterisk(*) after `yield`?');
            }
        }
        catch (e) {
            if (e.$$RESOLVE === true) {
                return resumeTask(e.$$VALUE);
            }
            else if (e.$$REJECT === true) {
                return throwTask(e.$$VALUE);
            }
            else {
                return throwTask(e);
            }
        }
    }
    function resumeTask(resumeValue) {
        currentDisposer = null;
        let result;
        try {
            result = iter.next(resumeValue);
        }
        catch (e) {
            return throwTask(e);
        }
        if (result.done) {
            settled || ((settled = true), onsuccess(result.value));
            return;
        }
        else {
            takeControl(result.value);
        }
    }
    function throwTask(error) {
        currentDisposer = null;
        let result;
        try {
            result = iter.throw(error);
        }
        catch (e) {
            if (error === e) {
                settled || ((settled = true), onfailure(e));
                return;
            }
            else {
                return throwTask(e);
            }
        }
        if (result.done) {
            settled || ((settled = true), onsuccess(result.value));
            return;
        }
        takeControl(result.value);
    }
    resumeTask(undefined); //start synchronously.
    return () => {
        if (!settled) {
            try {
                currentDisposer?.();
            }
            catch (e) {
                if (!(e instanceof CanceledError)) {
                    throw e;
                }
            }
            finally {
                settled = true;
            }
        }
    };
}
function executeAsyncRunnable(runnable, onsuccess, onfailure) {
    let currentDisposer = null;
    const iter = runnable[Symbol.asyncIterator]();
    let settled = false;
    function takeControl(yielded) {
        try {
            if (typeof yielded === 'function') {
                //bind sideEffect
                let synchronous = true; // closure callback
                let innerSettled = false;
                let synchronousResult = null;
                function tryResolve(value) {
                    if (innerSettled)
                        return;
                    innerSettled = true;
                    if (synchronous) {
                        synchronousResult = {
                            $$RESOLVE: true,
                            $$VALUE: value,
                        };
                    }
                    else {
                        resumeTask(value);
                    }
                }
                function tryReject(error) {
                    if (innerSettled)
                        return;
                    innerSettled = true;
                    if (synchronous) {
                        synchronousResult = {
                            $$REJECT: true,
                            $$VALUE: error,
                        };
                    }
                    else {
                        throwTask(error);
                    }
                }
                const maybeDisposer = yielded(tryResolve, tryReject); //might raise an error.
                if (synchronousResult !== null) {
                    throw synchronousResult;
                }
                currentDisposer = () => {
                    // if (innerSettled) return; // no need this line: currentDisposer
                    // is guaranteed to execute only once.
                    doCleanup(maybeDisposer); // maybe raise an error.
                    innerSettled || tryReject(new CanceledError());
                    // innerSettled = true;
                };
                synchronous = false;
            }
            else {
                throw new TypeError('Invalid object yielded. Are you missing an asterisk(*) after `yield`?');
            }
        }
        catch (e) {
            if (e.$$RESOLVE === true) {
                return resumeTask(e.$$VALUE);
            }
            else if (e.$$REJECT === true) {
                return throwTask(e.$$VALUE);
            }
            else {
                return throwTask(e);
            }
        }
    }
    async function resumeTask(resumeValue) {
        currentDisposer = null;
        let result;
        try {
            result = await iter.next(resumeValue);
        }
        catch (e) {
            return throwTask(e);
        }
        if (result.done) {
            settled || ((settled = true), onsuccess(result.value));
            return;
        }
        else {
            takeControl(result.value);
        }
    }
    async function throwTask(error) {
        currentDisposer = null;
        let result;
        try {
            result = await iter.throw(error);
        }
        catch (e) {
            if (error === e) {
                settled || ((settled = true), onfailure(e));
                return;
            }
            else {
                return throwTask(e);
            }
        }
        if (result.done) {
            settled || ((settled = true), onsuccess(result.value));
            return;
        }
        takeControl(result.value);
    }
    resumeTask(undefined); //start synchronously.
    return () => {
        if (!settled) {
            try {
                currentDisposer?.();
            }
            catch (e) {
                if (!(e instanceof CanceledError)) {
                    throw e;
                }
            }
            finally {
                settled = true;
            }
        }
    };
}
// bug: type inference breaks when calling itself.
function task(taskFn) {
    return function (...params) {
        return new CancellablePromise((resolve, reject) => {
            return executeRunnable(taskFn.call(undefined, ...params), resolve, reject);
        });
    };
}
function start(runnable) {
    return new CancellablePromise((resolve, reject) => {
        return executeRunnable(runnable, resolve, reject);
    });
}
function* delay(time) {
    return (yield (resume) => {
        const id = setTimeout((s) => resume(s), time);
        return () => clearTimeout(id);
    });
}
function* timeout(time) {
    return (yield (_, reject) => {
        const id = setTimeout((s) => reject(new Error('Timeout')), time);
        return () => clearTimeout(id);
    });
}
/* istanbul ignore next */
function* nextAnimationFrame() {
    return (yield (resume) => {
        const id = requestAnimationFrame(resume);
        return () => cancelAnimationFrame(id);
    });
}
class Semaphore {
    constructor(maxConcurrency) {
        this.maxConcurrency = maxConcurrency;
        this.currentNum = 0;
        this.queue = [];
    }
    get free() {
        return this.currentNum < this.maxConcurrency;
    }
    *waitOne() {
        if (this.currentNum >= this.maxConcurrency) {
            yield* new CancellablePromise((resolve) => {
                this.queue.push(resolve);
                return () => this.queue.splice(this.queue.indexOf(resolve), 1); // resolve definitely exist.
            });
            // a free space is gurantted.
        }
        this.currentNum++;
        return;
    }
    // this is an action
    release() {
        if (this.currentNum === 0) {
            return;
        }
        this.currentNum--;
        // trigger?
        if (this.queue.length) {
            this.queue.shift()(); // a microtask is scheduled.
        }
    }
}
function* resolve(obj) {
    if (typeof obj === 'object' && obj !== null) {
        if (Symbol.iterator in obj) {
            return yield* obj;
        }
        else if ('then' in obj) {
            return yield (resolve, reject) => {
                obj.then(resolve, reject);
                return noop;
            };
        }
        else if (Symbol_observable in obj) {
            return yield (resolve, reject) => {
                let lastEmit = undefined;
                const subscription = obj[Symbol_observable]().subscribe((next) => {
                    lastEmit = next;
                }, (error) => reject(error), () => {
                    resolve(lastEmit);
                });
                return () => subscription.unsubscribe();
            };
        }
        else {
            return yield (_, reject) => {
                reject(new TypeError("Can't resolve object"));
                return noop;
            };
        }
    }
    else {
        return yield (resolve) => {
            resolve(obj);
            return noop;
        };
    }
}
function* any(tasks) {
    return yield (success, fail) => {
        let count = 0;
        const errors = [];
        const disposers = [];
        let sync = true;
        let errorCount = 0;
        for (const task of tasks) {
            const currentIndex = count;
            try {
                disposers.push(executeRunnable(resolve(task), (value) => {
                    if (sync)
                        throw value;
                    success(value);
                    disposers.forEach((x) => x()); // async
                }, (error) => {
                    errorCount++;
                    errors[currentIndex] = error;
                    if (!sync) {
                        if (errorCount === count) {
                            fail(new AggregateError(errors, 'All task failed'));
                        }
                    }
                }));
            }
            catch (value) {
                success(value);
                disposers.forEach((x) => x()); // sync
                return;
            }
            count++;
        }
        sync = false;
        if (errorCount === count) {
            return fail(new AggregateError(errors, 'All task failed'));
        }
        return () => {
            fail(new CanceledError());
            disposers.forEach((x) => x());
        };
    };
}
function* race(tasks) {
    return yield (success, fail) => {
        const disposers = [];
        let sync = true;
        for (const task of tasks) {
            try {
                disposers.push(executeRunnable(resolve(task), (value) => {
                    if (sync) {
                        throw {
                            $$RESOLVE: true,
                            $$VALUE: value,
                        };
                    }
                    success(value);
                    disposers.forEach((x) => x()); // async
                }, (error) => {
                    if (sync)
                        throw {
                            $$REJECT: true,
                            $$VALUE: error,
                        };
                    fail(error);
                    disposers.forEach((x) => x()); // async
                }));
            }
            catch (value) {
                if (value.$$RESOLVE === true) {
                    success(value.$$VALUE);
                }
                else if (value.$$REJECT === true) {
                    fail(value.$$VALUE);
                }
                else {
                    fail(value); // highly impossible, but for insurance
                }
                sync = false; // NB
                disposers.forEach((x) => x()); // sync
                return;
            }
        }
        sync = false;
        return () => {
            fail(new CanceledError());
            disposers.forEach((x) => x());
        };
    };
}
// it's the opposite of any!
function* all(tasks) {
    return yield (success, fail) => {
        let count = 0;
        const result = [];
        const disposers = [];
        let sync = true;
        let succeedCount = 0;
        for (const task of tasks) {
            const currentIndex = count;
            try {
                disposers.push(executeRunnable(resolve(task), (value) => {
                    succeedCount++;
                    result[currentIndex] = value;
                    if (!sync) {
                        if (succeedCount === count) {
                            success(result);
                        }
                    }
                }, (error) => {
                    if (sync)
                        throw error;
                    fail(error);
                    disposers.forEach((x) => x()); //async
                }));
            }
            catch (e) {
                fail(e);
                sync = false;
                disposers.forEach((x) => x()); //sync
                return;
            }
            count++;
        }
        sync = false;
        if (succeedCount === count) {
            return success(result);
        }
        return () => {
            fail(new CanceledError());
            disposers.forEach((x) => x());
        };
    };
}
function* allSettled(tasks) {
    return yield (success, fail) => {
        let count = 0;
        const result = [];
        const disposers = [];
        let sync = true;
        let settledCount = 0;
        for (const task of tasks) {
            const currentIndex = count;
            try {
                disposers.push(executeRunnable(resolve(task), (value) => {
                    settledCount++;
                    result[currentIndex] = {
                        success: true,
                        value,
                    };
                    if (!sync) {
                        if (settledCount === count) {
                            success(result);
                        }
                    }
                }, (error) => {
                    settledCount++;
                    result[currentIndex] = {
                        success: false,
                        value: error,
                    };
                    if (!sync) {
                        if (settledCount === count) {
                            success(result);
                        }
                    }
                }));
            }
            catch (e) {
                // almost you can catch nothing, but for insurance
                fail(e);
                disposers.forEach((x) => x());
                return;
            }
            count++;
        }
        sync = false;
        if (settledCount === count) {
            return success(result);
        }
        return () => {
            fail(new CanceledError());
            disposers.forEach((x) => x());
        };
    };
}

function filterCanceledError(x) {
    if (!(x instanceof CanceledError)) {
        throw x;
    }
}
let currentCollecting = null;
function collectScope() {
    const stored = currentCollecting;
    currentCollecting = [];
    const current = currentCollecting;
    return () => {
        if (currentCollecting !== current) {
            throw new Error('Nested scope');
        }
        currentCollecting = stored;
        return new LifecycleScope(current);
    };
}
class LifecycleScope {
    constructor(onmountLogics) {
        this.onmountLogics = onmountLogics;
        this.attached = false;
        this.detach = noop;
    }
    attach() {
        /* istanbul ignore if */
        if (this.attached) {
            return;
        }
        this.attached = true;
        const cleanups = this.onmountLogics.map((x) => {
            const cleanup = x();
            if (cleanup instanceof CancellablePromise) {
                // TODO: ignore specific error: CancelledError
                cleanup.catch(filterCanceledError);
            }
            return cleanup;
        });
        return (this.detach = () => {
            if (this.attached) {
                this.attached = false;
                cleanups.forEach(doCleanup);
            }
        });
    }
}
function lifecycle(onInit) {
    if (currentCollecting) {
        currentCollecting.push(onInit);
        return;
    }
    throw new TypeError('Not inside a scope');
}

const Identifier = {
    of: (description, options) => {
        const id = Symbol(description);
        if (options) {
            if (options.suggesstion) ;
        }
        return id;
    },
};
let currentContext = null;
class Context {
    constructor(entries = new Map()) {
        this.entries = entries;
    }
    get(token, options) {
        if (this.entries.has(token)) {
            return this.entries.get(token);
        }
        if (options?.optional) {
            return options.defaultValue ?? undefined;
        }
        throw new IdentifierNotFoundError(token);
    }
    runInContext() {
        const stored = currentContext;
        currentContext = this;
        return () => {
            if (currentContext !== this) {
                throw new Error('Exit inner context before outer context.');
            }
            currentContext = stored;
        };
    }
    inherit(module) {
        return new Context(new Map([
            ...this.entries,
            ...Object.getOwnPropertySymbols(module).map((ssymbol) => [ssymbol, module[ssymbol]]),
        ]));
    }
    /**
     * Build a new context based on current from concern.
     * @param concern
     * @returns
     */
    build(concern) {
        const exitContext = this.runInContext();
        try {
            const exports = concern() ?? {};
            return this.inherit(exports);
        }
        finally {
            exitContext();
        }
    }
}
Context.EMPTY = new Context();
class IdentifierNotFoundError extends Error {
    constructor(identifier) {
        super(`'${identifier.description}' is not found in current context.`);
        this.identifier = identifier;
    }
}
function injected(token, options) {
    if (currentContext) {
        return currentContext.get(token, options);
    }
    throw new TypeError('Not inside a context.');
}
function reduceConcerns(concerns) {
    return () => {
        const exports = {};
        let context = currentContext; // TODO: maybe not?
        for (const concernOrArray of concerns) {
            const concern = concernOrArray instanceof Array
                ? reduceConcerns(concernOrArray)
                : concernOrArray;
            const exitContext = context.runInContext();
            try {
                const concernExport = concern() ?? {};
                Object.assign(exports, concernExport);
                context = context.inherit(concernExport);
            }
            catch (e) {
                if (e instanceof IdentifierNotFoundError) {
                    throw new ReferenceError(`Concern [${concern.name}] requires "${e.identifier.description}" but it's not defined.`);
                }
                throw e;
            }
            finally {
                exitContext();
            }
        }
        return exports;
    };
}

function held(stream, initial) {
    const [behavior, setBehavior] = mutValue(initial);
    lifecycle(() => stream.listen(setBehavior));
    return behavior;
}
function reduced(stream, reducer, initial) {
    const [behavior, setBehavior] = mutValue(initial);
    lifecycle(() => stream.listen((next) => {
        setBehavior(reducer(behavior.value, next));
    }));
    return behavior;
}

function debounce(time) {
    let timeoutId = undefined; // number or NodeJS.Timeout
    return (next) => {
        return (payload) => {
            if (timeoutId !== undefined) {
                clearTimeout(timeoutId);
            }
            timeoutId = setTimeout(() => {
                next(payload);
                timeoutId = undefined;
            }, time);
        };
    };
}
function throttle(time) {
    let timeoutId = undefined;
    let lastFired = 0;
    return (next) => {
        return (payload) => {
            if (Date.now() - lastFired > time) {
                next(payload);
                lastFired = Date.now();
            }
            else {
                if (timeoutId !== undefined) {
                    clearTimeout(timeoutId);
                }
                timeoutId = setTimeout(() => {
                    next(payload);
                    lastFired = Date.now();
                    timeoutId = undefined;
                }, Date.now() - lastFired);
            }
        };
    };
}
const animation = (next) => {
    let rafId = 0;
    return (payload) => {
        // last rafId is auto disposed
        rafId = requestAnimationFrame(() => {
            next(payload);
            rafId = 0;
        });
        return () => cancelAnimationFrame(rafId);
    };
};
function scheduled(source, scheduler) {
    const [dest, emitDest] = stream();
    lifecycle(() => source.listen(scheduler(emitDest)));
    return dest;
}

class Reference {
    constructor(current = null) {
        this.current = current;
    }
    get bind() {
        return (x) => {
            this.current = x;
        };
    }
    set bind(x) {
        this.current = x;
    }
}
function reference(initialValue) {
    return new Reference(initialValue);
}

function testBed(setup, concern) {
    const exitScope = collectScope();
    concern && new Context().build(concern);
    const binding = setup(createInteractionLogic(exitScope));
    return binding;
}
function createInteractionLogic(exitScope) {
    return function interactFn(interactLogic) {
        return {
            expectEffects(runnable) {
                const scope = exitScope();
                const detach = scope.attach();
                const expect = start(runnable()); // do side effects first
                start(interactLogic()); // start interact
                return expect.finally(detach);
            },
        };
    };
}

function effect(sideEffect) {
    const doEffect = () => {
        try {
            sideEffect();
        }
        catch (e) {
            console.error(`Captured error inside effect().`, e);
        }
    };
    const reaction = new Reaction(() => {
        reaction.execute(doEffect);
    });
    lifecycle(() => {
        reaction.execute(doEffect);
        return () => reaction.dispose();
    });
}
// export class LayoutQueue {
//   private queue: (() => void)[] = [];
//   enqueue(job: () => void) {
//     if (this.queue.length === 0) {
//       requestAnimationFrame(() => {
//         if (!this.willFlush /* template will take care */ && this.queue.length) {
//           this.flush();
//         }
//       });
//     }
//     this.queue.push(job);
//   }
//   private willFlush: boolean = false;
//   preflush() {
//     this.willFlush = true;
//   }
//   flush() {
//     this.willFlush = false;
//     try {
//       for (const job of this.queue) {
//         job();
//       }
//     } finally {
//       this.queue.length = 0;
//     }
//   }
// }
// export const LAYOUT_QUEUE = Identifier.of<LayoutQueue>(LayoutQueue.name);
// export function layout(sideEffect: () => void) {
//   const queue = consume(LAYOUT_QUEUE);
//   const doEffect = () => {
//     try {
//       sideEffect();
//     } catch (e) {
//       console.error(`Captured error inside layout().`, e);
//     }
//   };
//   const reaction = new Reaction(() => {
//     queue.enqueue(() => reaction.execute(doEffect));
//   });
//   lifecycle(() => {
//     reaction.execute(doEffect);
//     return () => reaction.dispose();
//   });
// }

function isCell(value) {
    return value instanceof Cell;
}
function isEventStream(value) {
    return value instanceof EventStream;
}

exports.CanceledError = CanceledError;
exports.CancellablePromise = CancellablePromise;
exports.Cell = Cell;
exports.Context = Context;
exports.EventStream = EventStream;
exports.Identifier = Identifier;
exports.IdentifierNotFoundError = IdentifierNotFoundError;
exports.LifecycleScope = LifecycleScope;
exports.Reaction = Reaction;
exports.Reference = Reference;
exports.Semaphore = Semaphore;
exports.SuspendedCell = SuspendedCell;
exports.Symbol_observable = Symbol_observable;
exports.__current_collecting = __current_collecting;
exports.all = all;
exports.allSettled = allSettled;
exports.animation = animation;
exports.any = any;
exports.batch = batch;
exports.collectScope = collectScope;
exports.computed = computed;
exports.debounce = debounce;
exports.delay = delay;
exports.effect = effect;
exports.executeAsyncRunnable = executeAsyncRunnable;
exports.executeRunnable = executeRunnable;
exports.held = held;
exports.injected = injected;
exports.isCell = isCell;
exports.isEventStream = isEventStream;
exports.lifecycle = lifecycle;
exports.merged = merged;
exports.mut = mutable;
exports.mutArray = mutArray;
exports.mutMap = mutMap;
exports.mutSet = mutSet;
exports.mutValue = mutValue;
exports.mutable = mutable;
exports.never = never;
exports.nextAnimationFrame = nextAnimationFrame;
exports.race = race;
exports.readEvents = readEvents;
exports.reduceConcerns = reduceConcerns;
exports.reduced = reduced;
exports.reference = reference;
exports.resolve = resolve;
exports.scheduled = scheduled;
exports.start = start;
exports.stream = stream;
exports.suspended = suspended;
exports.task = task;
exports.testBed = testBed;
exports.throttle = throttle;
exports.timeout = timeout;
